package fr.emse.gitlab.saref.jobs;

import java.io.File;
import java.io.FileReader;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.apache.jena.datatypes.xsd.XSDDatatype;
import org.apache.jena.query.Dataset;
import org.apache.jena.query.QueryExecution;
import org.apache.jena.query.QueryExecutionFactory;
import org.apache.jena.query.ReadWrite;
import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.rdf.model.Resource;
import org.apache.jena.rdf.model.ResourceFactory;
import org.apache.jena.tdb.TDBFactory;
import org.apache.jena.vocabulary.DCTerms;
import org.apache.jena.vocabulary.OWL2;
import org.apache.jena.vocabulary.RDF;
import org.apache.jena.vocabulary.RDFS;
import org.eclipse.jgit.api.Git;

import fr.emse.gitlab.saref.Constants;
import fr.emse.gitlab.saref.entities.git.Repositories;
import fr.emse.gitlab.saref.entities.git.Repository;
import fr.emse.gitlab.saref.entities.git.Version;
import fr.emse.gitlab.saref.vocabs.EX;
import fr.emse.gitlab.saref.vocabs.RDFP;
import fr.emse.gitlab.saref.vocabs.VANN;

public class CreateDataset extends AbstractJobRunner<Dataset> {

	private static final SimpleDateFormat SIMPLE_DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd");
	private static final Resource ETSI_URL = ResourceFactory.createResource("https://www.etsi.org/");
	private static final Resource ETSI_LICENSE = ResourceFactory
			.createResource("https://forge.etsi.org/etsi-software-license");

	private final Repositories repositories;
	private final Dataset dataset;
	private final Model config = ModelFactory.createDefaultModel();
	private final Map<String, Model> ngs = new HashMap<>();

	public CreateDataset(File directory, Repositories repositories) {
		super("Create the RDF Dataset", directory);
		this.repositories = repositories;
		dataset = TDBFactory.createDataset(Constants.DATASET_DIR);
	}

	@Override
	protected Dataset getValue() {
		for (Repository repository : repositories) {
			readRepository(repository);
		}
		writeToTDBDataset();
		return dataset;
	}

	private void readRepository(Repository repo) {
		try (Git git = Git.open(repo.getDirectory())) {
			config.add(repo.getResource(), RDF.type, OWL2.Ontology);
			for (int i = 0; i < repo.getVersions().size(); i++) {
				readVersion(repo, git, i);
			}
		} catch (Exception ex) {
			failure("Exception while reading repository " + repo.getName(), ex);
		}
	}

	private void readVersion(Repository repo, Git git, int i) {
		Version version = repo.getVersions().get(i);
		try {
			git.checkout().setName(version.getRef().getName()).call();
			Model onto = ModelFactory.createDefaultModel();
			File ontologyFile = new File(repo.getDirectory(), "ontology/" + repo.getOntologyFileName());
			try (FileReader fr = new FileReader(ontologyFile)) {
				onto.read(fr, Constants.BASE, "TTL");
			}
			ngs.put(version.getUri(), onto);
			addOntologyMetadata(onto, repo, git, i);
			augmentTermGraphs(onto, repo, git, i);
		} catch (Exception ex) {
			failure(String.format("Failed to read repository %s version %s", repo.getName(), version.getName()), ex);
		}
	}

	/**
	 * for each term, add: some triples in the default graph (used to detect naming
	 * clashes), and some triples in the named graph of the term (used to display
	 * the page of the term)
	 * 
	 * @param onto
	 * @param repo
	 * @param git
	 * @param i
	 */
	private void augmentTermGraphs(Model onto, Repository repo, Git git, int i) {
		Version version = repo.getVersions().get(i);
		Version priorVersion = i == 0 ? null : repo.getVersions().get(i - 1);
		Set<Resource> definedTerms = new HashSet<>();
		Set<Resource> usedTerms = new HashSet<>();
		computeTerms(onto, repo, version, definedTerms, usedTerms);
		config.add(repo.getResource(), EX.hasVersion, version.getResource());
		config.add(version.getResource(), RDF.type, EX.OntologyVersion);
		config.add(version.getResource(), EX.versionInfo, version.getVersionInfo());
		if (priorVersion != null) {
			config.add(version.getResource(), EX.priorVersion, priorVersion.getResource());
		}
		for (Resource t : definedTerms) {
			Model termModel = getNamedModel(t);
			String localName = t.getURI().substring(repo.getNamespace().length());
			termModel.add(t, RDFS.isDefinedBy, version.getResource());
			// keep the most recent definition of the term
			if (i == repo.getVersions().size() - 1) {
				termModel.add(t, DCTerms.modified, SIMPLE_DATE_FORMAT.format(version.getIssued()), XSDDatatype.XSDdate);
				try (QueryExecution exec = QueryExecutionFactory.create("DESCRIBE <" + t.getURI() + ">", onto)) {
					termModel.add(exec.execDescribe());
				}
			}
			config.add(t, RDF.type, RDFP.Resource);
			config.add(t, EX.localName, localName);
			config.add(t, RDFS.isDefinedBy, version.getResource());
		}
		for (Resource t : usedTerms) {
			Model termModel = getNamedModel(t);
			termModel.add(t, EX.isUsedBy, version.getResource());
			config.add(t, RDF.type, RDFP.Resource);
			config.add(t, EX.isUsedBy, version.getResource());
		}

	}

	private void writeToTDBDataset() {
		dataset.begin(ReadWrite.WRITE);
		dataset.addNamedModel(EX.config, config);
		for (String name : ngs.keySet()) {
			dataset.addNamedModel(name, ngs.get(name));
		}
		dataset.commit();
	}

	private void addOntologyMetadata(Model onto, Repository repo, Git git, int i) throws Exception {
		Version version = repo.getVersions().get(i);
		Version priorVersion = i == 0 ? null : repo.getVersions().get(i - 1);
		Resource resource = repo.getResource();
		onto.add(resource, RDF.type, OWL2.Ontology);

		onto.removeAll(resource, OWL2.versionIRI, null);
		onto.add(resource, OWL2.versionIRI, version.getResource());

		onto.removeAll(resource, OWL2.versionInfo, null);
		onto.add(resource, OWL2.versionInfo, version.getVersionInfo());

		onto.removeAll(resource, OWL2.priorVersion, null);
		if (priorVersion != null) {
			onto.add(resource, OWL2.priorVersion, priorVersion.getResource());
		}

		onto.removeAll(resource, DCTerms.publisher, null);
		onto.add(resource, DCTerms.publisher, ETSI_URL);

		onto.removeAll(resource, DCTerms.license, null);
		onto.add(resource, DCTerms.license, ETSI_LICENSE);

//		// list of contributors
//		Set<Contributor> contributors = new HashSet<>();
//		for (Iterator<RevCommit> it = git.log().add(version.ref.getObjectId()).call().iterator(); it.hasNext();) {
//			RevCommit commit = it.next();
//			Contributor contributor = new Contributor(commit.getCommitterIdent());
//			contributors.add(contributor);
//		}
//		for (Contributor contributor : contributors) {
//			// git can only provide name + email.
//			// would need to maintain some list of mapping email -> url in the future.
//			Resource anon = onto.createResource();
//			onto.add(repo.resource, DCTerms.contributor, anon);
//			onto.add(anon, FOAF.name, contributor.getName());
//			onto.add(anon, FOAF.mbox, contributor.getEmailAddress());
//		}

		onto.removeAll(resource, VANN.preferredNamespacePrefix, null);
		onto.removeAll(resource, VANN.preferredNamespaceUri, null);
		onto.add(resource, VANN.preferredNamespacePrefix, repo.getPrefix());
		onto.add(resource, VANN.preferredNamespaceUri, repo.getNamespace(), XSDDatatype.XSDanyURI);

		onto.removeAll(resource, DCTerms.issued, null);
		onto.add(resource, DCTerms.issued, SIMPLE_DATE_FORMAT.format(version.getIssued()), XSDDatatype.XSDdate);
	}

	private void computeTerms(Model onto, Repository repo, Version version, Set<Resource> definedTerms,
			Set<Resource> usedTerms) {
		onto.listStatements().forEachRemaining(stmt -> {
			Resource s = stmt.getSubject();
			Resource p = stmt.getPredicate();
			Resource o = stmt.getObject().isResource() ? (Resource) stmt.getObject() : null;
			computeTerms(s, repo, version, definedTerms, usedTerms);
			computeTerms(p, repo, version, definedTerms, usedTerms);
			computeTerms(o, repo, version, definedTerms, usedTerms);
		});
	}

	private void computeTerms(Resource t, Repository repo, Version version, Set<Resource> definedTerms,
			Set<Resource> usedTerms) {
		if (t == null || !t.isURIResource() || !t.getURI().startsWith(Constants.BASE) || t.getURI().endsWith("/")) {
			return;
		}
		String uri = t.getURI();
		if (uri.startsWith(repo.getNamespace())) {
			definedTerms.add(t);
		} else if (!uri.startsWith(repo.getNamespace())) {
			usedTerms.add(t);
		}
	}

	private Model getNamedModel(Resource t) {
		String uri = t.getURI();
		if (ngs.containsKey(uri)) {
			return ngs.get(uri);
		} else {
			Model model = ModelFactory.createDefaultModel();
			model.add(t, RDF.type, RDFS.Resource);
			ngs.put(uri, model);
			return model;
		}
	}

}
